#include <GL/glut.h>
#include <math.h>
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////////
//
//  GLOBALS
//
//////////////////////////////////////////////////////////////////////////

// math constants
static const float PI       = 3.14159f;
static const float TWO_PI   = 6.28318f;

// display list constants
static const int CUBE       = 1;
static const int SPHERE     = 2;

// angular velocities of the scene objects
static const float CUBE_ANGULAR_VELOCITY    = 0.25f;
static const float SPHERE_ANGULAR_VELOCITY  = 0.4f;

// rotation variables
static float g_cubeAngle                = 0.0f;
static float g_cubeAngularVelocity      = 0.0f;
static float g_sphereAngle              = 0.0f;
static float g_sphereAngularVelocity    = 0.0f;

// sphere data
static const int SPHERE_DIMENSIONS      = 17 * 17;

class CSphere
{
public:
    float vertices[SPHERE_DIMENSIONS][3];
    float colors[SPHERE_DIMENSIONS][3];
    float radius;
    int numSlices;
    int numStacks;
};

static CSphere g_sphere;

//////////////////////////////////////////////////////////////////////////
//
//  FUNCTIONS
//
//////////////////////////////////////////////////////////////////////////

// generate a random number in the range [0, 1)
inline float UnitRand()
{
    return rand() / ((float) RAND_MAX);
}


// initialize colors of sphere faces
void InitSphereColors()
{
    for ( int i = 0 ; i < SPHERE_DIMENSIONS ; ++i )
    {
        g_sphere.colors[i][0] = UnitRand();     // red component
        g_sphere.colors[i][1] = UnitRand();     // green component
        g_sphere.colors[i][2] = UnitRand();     // blue component
    }
}


void DrawSphere()
{
    int upperLeft,
        upperRight,
        lowerRight,
        lowerLeft;

    int jPlusOne;

    glBegin( GL_QUADS );
    
    for ( int i = 0 ; i < g_sphere.numStacks; ++i )
    {
        for ( int j = 0 ; j < g_sphere.numSlices ; ++j )
        {
            jPlusOne = j + 1;
            if ( jPlusOne == g_sphere.numSlices )
                jPlusOne = 0;

            upperLeft = i * g_sphere.numSlices + j;
            upperRight = i * g_sphere.numSlices + jPlusOne;
            lowerRight = (i + 1) * g_sphere.numSlices + jPlusOne;
            lowerLeft = (i + 1) * g_sphere.numSlices + j;

            glColor3fv( g_sphere.colors[i * g_sphere.numSlices + j] );
            glVertex3fv( g_sphere.vertices[upperLeft] );
            glVertex3fv( g_sphere.vertices[upperRight] );
            glVertex3fv( g_sphere.vertices[lowerRight] );
            glVertex3fv( g_sphere.vertices[lowerLeft] );
        }
    }

    glEnd();
}


// initialize sphere vertices
void InitSphereVertices(float radius, int numSlices, int numStacks)
{
    g_sphere.numSlices = numSlices;
    g_sphere.numStacks = numStacks;

    float theta,            // azimuthal angle
          phi;              // polar angle

    phi = 0.0f;
    for ( int i = 0 ; i < g_sphere.numStacks + 1 ; ++i )
    {
        theta = 0.0f;
        for ( int j = 0 ; j < g_sphere.numSlices ; ++j )
        {
            const int index = i * g_sphere.numSlices + j;
            g_sphere.vertices[index][0] = radius * sin( phi ) * sin( theta );
            g_sphere.vertices[index][1] = radius * cos( phi );
            g_sphere.vertices[index][2] = radius * sin( phi ) * cos( theta );

            theta += TWO_PI / g_sphere.numSlices;
        }

        phi += PI / g_sphere.numStacks;
    }

    // compile a display list for the sphere
    glNewList( SPHERE, GL_COMPILE );
      DrawSphere();
    glEndList();
}


void InitializeCube()
{
    static float cubeData[][3] = { {1, 1, 1},
                                   {-1, 1, 1},
                                   {-1, -1, 1},
                                   {1, -1, 1},
                                   {1, 1, -1},
                                   {-1, 1, -1},
                                   {-1, -1, -1},
                                   {1, -1, -1} };

    // compile a display list for the cube
    glNewList( CUBE, GL_COMPILE );
        glBegin( GL_QUADS );
            // front face
            glColor3f( 1, 0, 0 );
            glVertex3fv( cubeData[0] );
            glVertex3fv( cubeData[1] );
            glVertex3fv( cubeData[2] );
            glVertex3fv( cubeData[3] );

            // back face
            glColor3f( 0, 1, 0 );
            glVertex3fv( cubeData[4] );
            glVertex3fv( cubeData[5] );
            glVertex3fv( cubeData[6] );
            glVertex3fv( cubeData[7] );

            // top face
            glColor3f( 0, 0, 1 );
            glVertex3fv( cubeData[0] );
            glVertex3fv( cubeData[1] );
            glVertex3fv( cubeData[5] );
            glVertex3fv( cubeData[4] );

            // bottom face
            glColor3f( 1, 1, 0 );
            glVertex3fv( cubeData[2] );
            glVertex3fv( cubeData[3] );
            glVertex3fv( cubeData[7] );
            glVertex3fv( cubeData[6] );

            // left face
            glColor3f( 1, 0, 1 );
            glVertex3fv( cubeData[1] );
            glVertex3fv( cubeData[2] );
            glVertex3fv( cubeData[6] );
            glVertex3fv( cubeData[5] );

            // right face
            glColor3f( 0, 1, 1 );
            glVertex3fv( cubeData[0] );
            glVertex3fv( cubeData[3] );
            glVertex3fv( cubeData[7] );
            glVertex3fv( cubeData[4] );
        glEnd();
    glEndList();
}


void Initialize()
{
    // initialize matrix stacks
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 60.0, 640.0 / 480.0, 1.0, 20.0 );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glTranslatef( 0, 0, -13 );

    // describe a viewport
    glViewport( 0, 0, 640, 480 );

    // enable Z-buffering
    glEnable( GL_DEPTH_TEST );

    // initialize data
    InitSphereColors();
    InitSphereVertices( 2.0f, 16, 16 );
    InitializeCube();
}


void DisplayCallback()
{
    // clear the color buffer
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // draw spheres
    glPushMatrix();
        glTranslatef( -5, 0, 0 );
        glRotatef( g_sphereAngle, 0.0f, 1.0f, 0.0f );
        glCallList( SPHERE );
    glPopMatrix();

    glPushMatrix();
        glRotatef( g_sphereAngle, 1.0f, 0.0f, 0.0f );
        glCallList( SPHERE );
    glPopMatrix();

    glPushMatrix();
        glTranslatef( 5, 0, 0 );
        glRotatef( g_sphereAngle, 0.0f, 0.0f, 1.0f );
        glCallList( SPHERE );
    glPopMatrix();

    // draw cubes
    glPushMatrix();
        glTranslatef( 0, 5, 0 );
        glRotatef( g_cubeAngle, 1.0f, 1.0f, 1.0f );
        glCallList( CUBE );
    glPopMatrix();

    glPushMatrix();
        glTranslatef( 0, -5, 0 );
        glRotatef( g_cubeAngle, -1.0f, 1.0f, 1.0f );
        glCallList( CUBE );
    glPopMatrix();

    // draw the buffer to the screen
    glutSwapBuffers();
}


void IdleCallback()
{
    g_cubeAngle += g_cubeAngularVelocity;
    g_sphereAngle += g_sphereAngularVelocity;

    glutPostRedisplay();
}


void KeyboardCallback(unsigned char key, int x, int y)
{
    switch( key )
    {
        // <SPACE> toggles object rotation
        case( ' ' ):
            if ( g_cubeAngularVelocity > 0.0f )
                g_cubeAngularVelocity = 0.0f;
            else
                g_cubeAngularVelocity = CUBE_ANGULAR_VELOCITY;

            if ( g_sphereAngularVelocity > 0.0f )
                g_sphereAngularVelocity = 0.0f;
            else
                g_sphereAngularVelocity = SPHERE_ANGULAR_VELOCITY;
            break;

        // <ESC> to quit
        case( 27 ):
            exit( 0 );
            break;
    }
}


int main(int argc, char **argv)
{
    // create window and rendering context
    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
    glutInitWindowSize( 640, 480 );
    glutCreateWindow( "animDemo" );

    // register callbacks
    glutDisplayFunc( DisplayCallback );
    glutIdleFunc( IdleCallback );
    glutKeyboardFunc( KeyboardCallback );

    // do some initialization
    Initialize();

    // pass control over to GLUT
    glutMainLoop();

    return 0;           // never reached
}
